package ru.dbotthepony.mc.otm.block.entity.tech

import net.minecraft.core.BlockPos
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.enchantment.Enchantments
import net.minecraft.world.level.block.state.BlockState
import ru.dbotthepony.mc.otm.block.entity.MatteryDeviceBlockEntity
import ru.dbotthepony.mc.otm.capability.CombinedItemHandler
import ru.dbotthepony.mc.otm.container.HandlerFilter
import ru.dbotthepony.mc.otm.container.MatteryContainer
import ru.dbotthepony.mc.otm.item.EssenceCapsuleItem
import ru.dbotthepony.mc.otm.menu.tech.EssenceStorageMenu
import ru.dbotthepony.mc.otm.registry.MBlockEntities

class EssenceStorageBlockEntity(blockPos: BlockPos, blockState: BlockState) : MatteryDeviceBlockEntity(MBlockEntities.ESSENCE_STORAGE, blockPos, blockState) {
	var experienceStored = 0L
		set(value) {
			require(value >= 0L) { "Negative experience: $value" }
			field = value
			markDirtyFast()
		}

	val capsuleContainer = MatteryContainer(::markDirtyFast, 1)
	val servoContainer = MatteryContainer(::markDirtyFast, 1)
	val mendingContainer = MatteryContainer(::markDirtyFast, 1).also(::addDroppableContainer)

	init {
		savetables.long(::experienceStored)
		savetables.stateful(::capsuleContainer)
		savetables.stateful(::servoContainer)
		savetables.stateful(::mendingContainer)
	}

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu {
		return EssenceStorageMenu(containerID, inventory, this)
	}

	val itemConfig = ConfigurableItemHandler(
		inputOutput = CombinedItemHandler(
			capsuleContainer.handler(HandlerFilter.OnlyIn.and(object : HandlerFilter {
				override fun canInsert(slot: Int, stack: ItemStack): Boolean {
					return stack.item is EssenceCapsuleItem
				}
			})),
			mendingContainer.handler(object : HandlerFilter {
				override fun canInsert(slot: Int, stack: ItemStack): Boolean {
					return stack.isDamaged && stack.getEnchantmentLevel(Enchantments.MENDING) > 0
				}

				override fun canExtract(slot: Int, amount: Int, stack: ItemStack): Boolean {
					return !stack.isDamaged || experienceStored <= 0
				}
			})
		)
	)

	override fun tick() {
		super.tick()

		if (!redstoneControl.isBlockedByRedstone) {
			val capsule = capsuleContainer[0]

			if (!capsule.isEmpty && capsule.item is EssenceCapsuleItem) {
				experienceStored += EssenceCapsuleItem.experienceStored(capsule)
				capsuleContainer.clearContent()
			}

			val repairStack = mendingContainer[0]

			if (!repairStack.isEmpty && repairStack.isDamaged) {
				if (experienceStored > 0) {
					val dmgPerExp = repairStack.xpRepairRatio
					var repairPoints = dmgPerExp.toInt()

					val diff = dmgPerExp - repairPoints.toFloat()
					if (diff > 0f) {
						repairPoints += if ((level?.random?.nextFloat() ?: 1f) <= diff) 1 else 0
					}

					experienceStored -= 1
					repairStack.damageValue -= repairPoints
				}
			}
		}
	}
}
